==> mars-0.9.2.tar.gz: mars-0.9.2/src/host/lib/mars_task_event_flag.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <sched.h>
#include <string.h>

#include "mars/mars_task_event_flag.h"
#include "mars/mars_workload_queue.h"
#include "mars/mars.h"
#include "mars/mars_mutex.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

int mars_task_event_flag_initialize(struct mars_context *mars,
				struct mars_task_event_flag *event_flag,
				uint8_t direction,
				uint8_t clear_mode)
{
	MARS_CHECK_RET(mars, MARS_ERROR_NULL);
	MARS_CHECK_RET(event_flag, MARS_ERROR_NULL);
	MARS_CHECK_RET(((uintptr_t)event_flag & MARS_TASK_EVENT_FLAG_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(direction == MARS_TASK_EVENT_FLAG_HOST_TO_MPU ||
			direction == MARS_TASK_EVENT_FLAG_MPU_TO_HOST ||
			direction == MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET(clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO ||
			clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_MANUAL,
			MARS_ERROR_PARAMS);

	mars_mutex_initialize((struct mars_mutex *)event_flag);
	mars_mutex_lock((struct mars_mutex *)event_flag);

	event_flag->mars_context_ea = (uint64_t)(uintptr_t)mars;
	event_flag->bits = 0;
	event_flag->direction = direction;
	event_flag->clear_mode = clear_mode;
	event_flag->wait_count = 0;

	mars_mutex_unlock((struct mars_mutex *)event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_clear(struct mars_task_event_flag *event_flag,
				uint32_t bits)
{
	MARS_CHECK_RET(event_flag, MARS_ERROR_NULL);
	MARS_CHECK_RET(((uintptr_t)event_flag & MARS_TASK_EVENT_FLAG_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);

	mars_mutex_lock((struct mars_mutex *)event_flag);

	/* clear the necessary bits */
	event_flag->bits &= ~bits;

	mars_mutex_unlock((struct mars_mutex *)event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_set(struct mars_task_event_flag *event_flag,
				uint32_t bits)
{
	MARS_CHECK_RET(event_flag, MARS_ERROR_NULL);
	MARS_CHECK_RET(event_flag->mars_context_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET(event_flag->direction ==
			MARS_TASK_EVENT_FLAG_HOST_TO_MPU, MARS_ERROR_STATE);
	MARS_CHECK_RET(((uintptr_t)event_flag & MARS_TASK_EVENT_FLAG_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);

	int ret;
	int i;
	struct mars_context *mars =
		(void *)(uintptr_t)event_flag->mars_context_ea;

	mars_mutex_lock((struct mars_mutex *)event_flag);

	/* set the necessary bits */
	event_flag->bits |= bits;

	/* save current set bits */
	bits = event_flag->bits;

	/* search through wait list for tasks to be signalled */
	for (i = 0; i < event_flag->wait_count; i++) {
		/* check condition based on wait mode */
		switch (event_flag->wait_mask_mode[i]) {
		case MARS_TASK_EVENT_FLAG_MASK_OR:
			if ((bits & event_flag->wait_mask[i]) == 0)
				continue;
			break;
		case MARS_TASK_EVENT_FLAG_MASK_AND:
			if ((bits & event_flag->wait_mask[i]) !=
				event_flag->wait_mask[i])
				continue;
			break;
		}

		/* clear event if clear mode is auto */
		if (event_flag->clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO)
			event_flag->bits &= ~event_flag->wait_mask[i];

		/* signal the task to go to ready state */
		ret = workload_queue_signal_send(mars->workload_queue,
			event_flag->wait_id[i]);
		MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

		/* flush id from wait list */
		event_flag->wait_count--;
		memmove(&event_flag->wait_id[i],
			&event_flag->wait_id[i + 1],
			sizeof(uint16_t) * (event_flag->wait_count - i));
		memmove(&event_flag->wait_mask[i],
			&event_flag->wait_mask[i + 1],
			sizeof(uint32_t) * (event_flag->wait_count - i));
		memmove(&event_flag->wait_mask_mode[i],
			&event_flag->wait_mask_mode[i + 1],
			sizeof(uint8_t) * (event_flag->wait_count - i));
		i--;
	}

	mars_mutex_unlock((struct mars_mutex *)event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_wait(struct mars_task_event_flag *event_flag,
				uint32_t mask,
				uint8_t mask_mode)
{
	MARS_CHECK_RET(event_flag, MARS_ERROR_NULL);
	MARS_CHECK_RET(event_flag->direction ==
			MARS_TASK_EVENT_FLAG_MPU_TO_HOST, MARS_ERROR_STATE);
	MARS_CHECK_RET(mask_mode == MARS_TASK_EVENT_FLAG_MASK_OR ||
			mask_mode == MARS_TASK_EVENT_FLAG_MASK_AND,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET(((uintptr_t)event_flag & MARS_TASK_EVENT_FLAG_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);

	mars_mutex_lock((struct mars_mutex *)event_flag);

	/* check condition based on wait mode */
	switch (mask_mode) {
	case MARS_TASK_EVENT_FLAG_MASK_OR:
		while ((event_flag->bits & mask) == 0) {
			mars_mutex_unlock((struct mars_mutex *)event_flag);
			while ((event_flag->bits & mask) == 0)
				sched_yield();
			mars_mutex_lock((struct mars_mutex *)event_flag);
		}
		break;
	case MARS_TASK_EVENT_FLAG_MASK_AND:
		while ((event_flag->bits & mask) != mask) {
			mars_mutex_unlock((struct mars_mutex *)event_flag);
			while ((event_flag->bits & mask) != mask)
				sched_yield();
			mars_mutex_lock((struct mars_mutex *)event_flag);
		}
		break;
	}

	/* clear event if clear mode is auto */
	if (event_flag->clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO)
		event_flag->bits &= ~mask;

	mars_mutex_unlock((struct mars_mutex *)event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_try_wait(struct mars_task_event_flag *event_flag,
				uint32_t mask,
				uint8_t mask_mode)
{
	MARS_CHECK_RET(event_flag, MARS_ERROR_NULL);
	MARS_CHECK_RET(event_flag->direction ==
			MARS_TASK_EVENT_FLAG_MPU_TO_HOST, MARS_ERROR_STATE);
	MARS_CHECK_RET(mask_mode == MARS_TASK_EVENT_FLAG_MASK_OR ||
			mask_mode == MARS_TASK_EVENT_FLAG_MASK_AND,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET(((uintptr_t)event_flag & MARS_TASK_EVENT_FLAG_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);

	mars_mutex_lock((struct mars_mutex *)event_flag);

	/* check condition based on wait mode */
	switch (mask_mode) {
	case MARS_TASK_EVENT_FLAG_MASK_OR:
		if ((event_flag->bits & mask) == 0) {
			mars_mutex_unlock((struct mars_mutex *)event_flag);
			return MARS_ERROR_BUSY;
		}
		break;
	case MARS_TASK_EVENT_FLAG_MASK_AND:
		if ((event_flag->bits & mask) != mask) {
			mars_mutex_unlock((struct mars_mutex *)event_flag);
			return MARS_ERROR_BUSY;
		}
		break;
	}

	/* clear event if clear mode is auto */
	if (event_flag->clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO)
		event_flag->bits &= ~mask;

	mars_mutex_unlock((struct mars_mutex *)event_flag);

	return MARS_SUCCESS;
}

==> mars-08.06.26.tar.gz: mars-08.06.26/src/host/lib/mars_task_event_flag.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include "mars/mars_task_event_flag.h"
#include "mars/mars_workload_queue.h"
#include "mars/mars.h"
#include "mars/mars_mutex.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

int mars_task_event_flag_initialize(struct mars_context *mars,
				struct mars_task_event_flag *event_flag,
				uint8_t direction,
				uint8_t clear_mode)
{
	MARS_CHECK_RET(mars, MARS_ERROR_NULL);
	MARS_CHECK_RET(event_flag, MARS_ERROR_NULL);
	MARS_CHECK_RET(((uintptr_t)event_flag & MARS_TASK_EVENT_FLAG_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);
	MARS_CHECK_RET(direction == MARS_TASK_EVENT_FLAG_HOST_TO_MPU ||
			direction == MARS_TASK_EVENT_FLAG_MPU_TO_HOST ||
			direction == MARS_TASK_EVENT_FLAG_MPU_TO_MPU,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET(clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO ||
			clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_MANUAL,
			MARS_ERROR_PARAMS);

	mars_mutex_initialize((struct mars_mutex *)event_flag);
	mars_mutex_lock((struct mars_mutex *)event_flag);

	event_flag->mars_context_ea = (uint64_t)(uintptr_t)mars;
	event_flag->bits = 0;
	event_flag->direction = direction;
	event_flag->clear_mode = clear_mode;
	event_flag->wait_count = 0;

	mars_mutex_unlock((struct mars_mutex *)event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_clear(struct mars_task_event_flag *event_flag,
				uint32_t bits)
{
	MARS_CHECK_RET(event_flag, MARS_ERROR_NULL);
	MARS_CHECK_RET(((uintptr_t)event_flag & MARS_TASK_EVENT_FLAG_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);

	mars_mutex_lock((struct mars_mutex *)event_flag);

	/* clear the necessary bits */
	event_flag->bits &= ~bits;

	mars_mutex_unlock((struct mars_mutex *)event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_set(struct mars_task_event_flag *event_flag,
				uint32_t bits)
{
	MARS_CHECK_RET(event_flag, MARS_ERROR_NULL);
	MARS_CHECK_RET(event_flag->mars_context_ea, MARS_ERROR_NULL);
	MARS_CHECK_RET(event_flag->direction ==
			MARS_TASK_EVENT_FLAG_HOST_TO_MPU, MARS_ERROR_STATE);
	MARS_CHECK_RET(((uintptr_t)event_flag & MARS_TASK_EVENT_FLAG_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);

	int ret;
	int i;
	struct mars_context *mars =
			(struct mars_context *)event_flag->mars_context_ea;

	mars_mutex_lock((struct mars_mutex *)event_flag);

	/* set the necessary bits */
	event_flag->bits |= bits;

	/* save current set bits */
	bits = event_flag->bits;

	/* search through wait list for tasks to be signalled */
	for (i = 0; i < event_flag->wait_count; i++) {
		/* check condition based on wait mode */
		switch (event_flag->wait_mask_mode[i]) {
		case MARS_TASK_EVENT_FLAG_MASK_OR:
			if ((bits | event_flag->wait_mask[i]) == 0)
				continue;
			break;
		case MARS_TASK_EVENT_FLAG_MASK_AND:
			if ((bits & event_flag->wait_mask[i]) !=
				event_flag->wait_mask[i])
				continue;
			break;
		}

		/* clear event if clear mode is auto */
		if (event_flag->clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO)
			event_flag->bits &= ~event_flag->wait_mask[i];

		/* signal the task to go to ready state */
		ret = workload_queue_signal_send(mars->workload_queue,
			event_flag->wait_id[i]);
		MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

		/* flush id from wait list */
		event_flag->wait_count--;
		memmove(&event_flag->wait_id[i],
			&event_flag->wait_id[i + 1],
			sizeof(uint16_t) * (event_flag->wait_count - i));
		memmove(&event_flag->wait_mask[i],
			&event_flag->wait_mask[i + 1],
			sizeof(uint32_t) * (event_flag->wait_count - i));
		memmove(&event_flag->wait_mask_mode[i],
			&event_flag->wait_mask_mode[i + 1],
			sizeof(uint8_t) * (event_flag->wait_count - i));
		i--;
	}

	mars_mutex_unlock((struct mars_mutex *)event_flag);

	return MARS_SUCCESS;
}

int mars_task_event_flag_wait(struct mars_task_event_flag *event_flag,
				uint32_t mask,
				uint8_t mask_mode)
{
	MARS_CHECK_RET(event_flag, MARS_ERROR_NULL);
	MARS_CHECK_RET(event_flag->direction ==
			MARS_TASK_EVENT_FLAG_MPU_TO_HOST, MARS_ERROR_STATE);
	MARS_CHECK_RET(mask_mode == MARS_TASK_EVENT_FLAG_MASK_OR ||
			mask_mode == MARS_TASK_EVENT_FLAG_MASK_AND,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET(((uintptr_t)event_flag & MARS_TASK_EVENT_FLAG_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);

	do {
		/* check condition based on wait mode */
		switch (mask_mode) {
		case MARS_TASK_EVENT_FLAG_MASK_OR:
			while ((*(volatile uint32_t *)
				&event_flag->bits | mask) == 0) {
			}
			break;
		case MARS_TASK_EVENT_FLAG_MASK_AND:
			while ((*(volatile uint32_t *)
				&event_flag->bits & mask) != mask) {
			}
			break;
		}

		mars_mutex_lock((struct mars_mutex *)event_flag);

		/* check condition again after locking mutex */
		switch (mask_mode) {
		case MARS_TASK_EVENT_FLAG_MASK_OR:
			if ((event_flag->bits | mask) != 0) {
				event_flag->bits &= ~mask;
				mars_mutex_unlock(
					(struct mars_mutex *)event_flag);
				return MARS_SUCCESS;
			}
			break;
		case MARS_TASK_EVENT_FLAG_MASK_AND:
			if ((event_flag->bits & mask) == mask) {
				event_flag->bits &= ~mask;
				mars_mutex_unlock(
					(struct mars_mutex *)event_flag);
				return MARS_SUCCESS;
			}
			break;
		}

		mars_mutex_unlock((struct mars_mutex *)event_flag);
	} while (1);
}

int mars_task_event_flag_try_wait(struct mars_task_event_flag *event_flag,
				uint32_t mask,
				uint8_t mask_mode)
{
	MARS_CHECK_RET(event_flag, MARS_ERROR_NULL);
	MARS_CHECK_RET(event_flag->direction ==
			MARS_TASK_EVENT_FLAG_MPU_TO_HOST, MARS_ERROR_STATE);
	MARS_CHECK_RET(mask_mode == MARS_TASK_EVENT_FLAG_MASK_OR ||
			mask_mode == MARS_TASK_EVENT_FLAG_MASK_AND,
			MARS_ERROR_PARAMS);
	MARS_CHECK_RET(((uintptr_t)event_flag & MARS_TASK_EVENT_FLAG_ALIGN_MASK)
			== 0, MARS_ERROR_ALIGN);

	mars_mutex_lock((struct mars_mutex *)event_flag);

	/* check condition based on wait mode */
	switch (mask_mode) {
	case MARS_TASK_EVENT_FLAG_MASK_OR:
		if ((event_flag->bits | mask) == 0) {
			mars_mutex_unlock((struct mars_mutex *)event_flag);
			return MARS_ERROR_BUSY;
		}
		break;
	case MARS_TASK_EVENT_FLAG_MASK_AND:
		if ((event_flag->bits & mask) != mask) {
			mars_mutex_unlock((struct mars_mutex *)event_flag);
			return MARS_ERROR_BUSY;
		}
		break;
	}

	/* clear event if clear mode is auto */
	if (event_flag->clear_mode == MARS_TASK_EVENT_FLAG_CLEAR_AUTO)
		event_flag->bits &= ~mask;

	mars_mutex_unlock((struct mars_mutex *)event_flag);

	return MARS_SUCCESS;
}
